//------------------------------------------------------------------------------
/*
    This file is part of rippled: https://github.com/ripple/rippled
    Copyright (c) 2012, 2013 Ripple Labs Inc.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose  with  or without fee is hereby granted, provided that the above
    copyright notice and this permission notice appear in all copies.

    THE  SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
    WITH  REGARD  TO  THIS  SOFTWARE  INCLUDING  ALL  IMPLIED  WARRANTIES  OF
    MERCHANTABILITY  AND  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
    ANY  SPECIAL ,  DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
    WHATSOEVER  RESULTING  FROM  LOSS  OF USE, DATA OR PROFITS, WHETHER IN AN
    ACTION  OF  CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
//==============================================================================

#ifndef RIPPLE_NODESTORE_TYPES_H_INCLUDED
#define RIPPLE_NODESTORE_TYPES_H_INCLUDED

#include <ripple/basics/BasicConfig.h>
#include <ripple/nodestore/NodeObject.h>
#include <vector>

namespace ripple {
namespace NodeStore {

enum {
    // This is only used to pre-allocate the array for
    // batch objects and does not affect the amount written.
    //
    batchWritePreallocationSize = 256,

    // This sets a limit on the maximum number of writes
    // in a batch. Actual usage can be twice this since
    // we have a new batch growing as we write the old.
    // The main goal is to avoid delays while persisting the ledger.
    //
    batchWriteLimitSize = 262144
};

/** Return codes from Backend operations. */
enum Status {
    ok,
    notFound,
    dataCorrupt,
    unknown,
    backendError,

    customCode = 100
};

/** A batch of NodeObjects to write at once. */
using Batch = std::vector<std::shared_ptr<NodeObject>>;

}  // namespace NodeStore

/** Shard states. */
enum class ShardState : std::uint32_t {
    acquire,     // Acquiring ledgers
    complete,    // Backend is ledger complete, database is unverified
    finalizing,  // Verifying database
    finalized,   // Database verified, shard is immutable
    queued       // Queued to be finalized
};

}  // namespace ripple

#endif
